# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock
from kafka.common import PartitionMetadata

from kafka_utils.kafka_check.commands.min_isr import get_min_isr
from kafka_utils.kafka_check.commands.min_isr import process_metadata_response


TOPICS_STATE = {
    'topic_0': {
        0: PartitionMetadata(
            topic='topic_0',
            partition=0,
            leader=170396635,
            replicas=(170396635, 170398981),
            isr=(170398981,),
            error=0,
        ),
    },
    'topic_1': {
        0: PartitionMetadata(
            topic='topic_1',
            partition=0,
            leader=170396635,
            replicas=(170396635, 170398981),
            isr=(170396635, 170398981),
            error=0,
        ),
    },
}


def test_get_min_isr_empty():
    TOPIC_CONFIG_WITHOUT_MIN_ISR = {'version': 1, 'config': {'retention.ms': '86400000'}}
    attrs = {'get_topic_config.return_value': TOPIC_CONFIG_WITHOUT_MIN_ISR}
    zk_mock = mock.MagicMock(**attrs)

    min_isr = get_min_isr(zk_mock, 'topic_0')

    zk_mock.get_topic_config.assert_called_once_with('topic_0')
    assert min_isr is None


def test_get_min_isr():
    TOPIC_CONFIG_WITH_MIN_ISR = {
        'version': 1,
        'config': {'retention.ms': '86400000', 'min.insync.replicas': '3'},
    }
    attrs = {'get_topic_config.return_value': TOPIC_CONFIG_WITH_MIN_ISR}
    zk_mock = mock.MagicMock(**attrs)

    min_isr = get_min_isr(zk_mock, 'topic_0')

    zk_mock.get_topic_config.assert_called_once_with('topic_0')
    assert min_isr == 3


def test_process_metadata_response_empty():
    result = process_metadata_response(
        topics={},
        zk=None,
        default_min_isr=None,
        verbose=False,
    )

    assert result == 0


@mock.patch(
    'kafka_utils.kafka_check.commands.min_isr.get_min_isr',
    return_value=1,
    autospec=True,
)
def test_run_command_all_ok(min_isr_mock):
    result = process_metadata_response(
        topics=TOPICS_STATE,
        zk=None,
        default_min_isr=None,
        verbose=False,
    )

    assert result == 0


@mock.patch(
    'kafka_utils.kafka_check.commands.min_isr.get_min_isr',
    return_value=None,
    autospec=True,
)
def test_run_command_all_ok_without_min_isr_in_zk(min_isr_mock):
    result = process_metadata_response(
        topics=TOPICS_STATE,
        zk=None,
        default_min_isr=None,
        verbose=False,
    )

    assert result == 0


@mock.patch(
    'kafka_utils.kafka_check.commands.min_isr.get_min_isr',
    return_value=None,
    autospec=True,
)
def test_run_command_with_default_min_isr(min_isr_mock):
    result = process_metadata_response(
        topics=TOPICS_STATE,
        zk=None,
        default_min_isr=1,
        verbose=False,
    )

    assert result == 0


@mock.patch(
    'kafka_utils.kafka_check.commands.min_isr.get_min_isr',
    return_value=None,
    autospec=True,
)
def test_run_command_with_fail_with_default_min_isr(min_isr_mock):
    result = process_metadata_response(
        topics=TOPICS_STATE,
        zk=None,
        default_min_isr=2,
        verbose=False,
    )

    assert result == 1


@mock.patch(
    'kafka_utils.kafka_check.commands.min_isr.get_min_isr',
    return_value=3,
    autospec=True,
)
def test_run_command_all_fail(min_isr_mock):
    result = process_metadata_response(
        topics=TOPICS_STATE,
        zk=None,
        default_min_isr=1,
        verbose=False,
    )

    assert result == 2



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from kafka_utils.kafka_cluster_manager.cluster_info.util import compute_optimum
from kafka_utils.kafka_cluster_manager.cluster_info.util import separate_groups


def test_compute_optimum():
    optimal, extra = compute_optimum(3, 10)

    assert optimal == 3
    assert extra == 1


def test_compute_optimum_zero_groups():
    with pytest.raises(ZeroDivisionError):
        optimal, extra = compute_optimum(0, 10)


def test_compute_optimum_zero_elements():
    optimal, extra = compute_optimum(10, 0)

    assert optimal == 0
    assert extra == 0


def test_separate_groups_balanced():
    groups = [4, 4, 4]
    total = 12

    overloaded, underloaded = separate_groups(groups, lambda x: x, total)

    assert not overloaded
    assert not underloaded


def test_separate_groups_almost_balanced():
    groups = [5, 5, 4]
    total = 14

    overloaded, underloaded = separate_groups(groups, lambda x: x, total)

    assert overloaded == [5, 5]
    assert underloaded == [4]


def test_separate_groups_unbalanced():
    groups = [4, 4, 3, 2]
    total = 13

    overloaded, underloaded = separate_groups(groups, lambda x: x, total)

    assert overloaded == [4, 4]
    assert underloaded == [2, 3]


def test_separate_groups_balanced_greater_total():
    groups = [4, 4, 4]
    total = 13

    overloaded, underloaded = separate_groups(groups, lambda x: x, total)

    assert underloaded == [4, 4, 4]


def test_separate_groups_balanced_much_greater_total():
    groups = [4, 4, 4]
    total = 20

    overloaded, underloaded = separate_groups(groups, lambda x: x, total)

    assert underloaded == [4, 4, 4]


def test_separate_groups_balanced_smaller_total():
    groups = [4, 4, 4]
    total = 6

    overloaded, underloaded = separate_groups(groups, lambda x: x, total)

    assert overloaded == [4, 4, 4]



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
from mock import sentinel

from kafka_utils.kafka_cluster_manager.cluster_info.partition import Partition


class TestPartition(object):

    @pytest.fixture
    def partition(self):
        mock_topic = sentinel.t1
        mock_topic.id = 't1'
        return Partition(
            mock_topic,
            0,
            [sentinel.r1, sentinel.r2],
        )

    def test_name(self, partition):
        assert partition.name == ('t1', 0)

    def test_topic(self, partition):
        assert partition.topic == sentinel.t1

    def test_replicas(self, partition):
        assert partition.replicas == [sentinel.r1, sentinel.r2]

    def test_leader(self, partition):
        assert partition.leader == sentinel.r1

    def test_replication_factor(self, partition):
        assert partition.replication_factor == 2

    def test_partition_id(self, partition):
        assert partition.partition_id == 0

    def test_add_replica(self, partition):
        new_broker = sentinel.new_r
        partition.add_replica(new_broker)
        assert partition.replicas == [sentinel.r1, sentinel.r2, sentinel.new_r]

    def test_swap_leader(self, partition):
        b = sentinel.r2
        old_replicas = partition.replicas
        partition.swap_leader(b)

        # Verify leader changed to b
        assert partition.leader == b
        # Verify that replica set remains same
        assert sorted(old_replicas) == sorted(partition.replicas)

    def test_followers_1(self, partition):
        # Case:1 With followers
        assert partition.followers == [sentinel.r2]

    def test_followers_2(self):
        # Case:2 No-followers
        mock_topic = sentinel.t1
        mock_topic.id = 't1'
        p2 = Partition(mock_topic, 0, [sentinel.r1])

        assert p2.followers == []

    def test_count_siblings(self):
        t1, t1.id = sentinel.t1, 't1'
        t2, t2.id = sentinel.t2, 't2'
        p1, p3, p4 = Partition(t1, 0), Partition(t1, 1), Partition(t2, 0),

        # verify sibling count
        p_group = [p1, p4, p3]
        assert p3.count_siblings(p_group) == 2
        assert p4.count_siblings(p_group) == 1
        p_group = [p4]
        assert p1.count_siblings(p_group) == 0

        # Empty group
        p_group = []
        assert p1.count_siblings(p_group) == 0

    def test_replace(self, partition):
        curr_broker = partition.replicas[0]
        partition.replace(curr_broker, sentinel.new_broker)

        assert partition.replicas[0] == sentinel.new_broker



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock
import pytest

from kafka_utils.kafka_consumer_manager. \
    commands.offset_restore import OffsetRestore
from kafka_utils.util.client import KafkaToolClient
from kafka_utils.util.monitoring import ConsumerPartitionOffsets


class TestOffsetRestore(object):

    topics_partitions = {
        "topic1": [0, 1, 2],
        "topic2": [0, 1, 2, 3],
        "topic3": [0, 1],
    }
    consumer_offsets_metadata = {
        'topic1':
        [
            ConsumerPartitionOffsets(topic='topic1', partition=0, current=20, highmark=655, lowmark=655),
            ConsumerPartitionOffsets(topic='topic1', partition=1, current=10, highmark=655, lowmark=655)
        ]
    }
    parsed_consumer_offsets = {'groupid': 'group1', 'offsets': {'topic1': {0: 10, 1: 20}}}
    new_consumer_offsets = {'topic1': {0: 10, 1: 20}}
    kafka_consumer_offsets = {'topic1': [
        ConsumerPartitionOffsets(topic='topic1', partition=0, current=30, highmark=40, lowmark=10),
        ConsumerPartitionOffsets(topic='topic1', partition=1, current=20, highmark=40, lowmark=10),
    ]}

    @pytest.fixture
    def mock_kafka_client(self):
        mock_kafka_client = mock.MagicMock(
            spec=KafkaToolClient
        )
        mock_kafka_client.get_partition_ids_for_topic. \
            side_effect = self.topics_partitions
        return mock_kafka_client

    def test_restore_offsets_zk(self, mock_kafka_client):
        with mock.patch(
            "kafka_utils.kafka_consumer_manager."
            "commands.offset_restore.set_consumer_offsets",
            return_value=[],
            autospec=True,
        ) as mock_set_offsets, mock.patch.object(
            OffsetRestore,
            "parse_consumer_offsets",
            spec=OffsetRestore.parse_consumer_offsets,
            return_value=self.parsed_consumer_offsets,
        ), mock.patch(
            "kafka_utils.kafka_consumer_manager."
            "commands.offset_restore.get_consumer_offsets_metadata",
            return_value=self.consumer_offsets_metadata,
            autospec=True,
        ):
            OffsetRestore.restore_offsets(
                mock_kafka_client,
                self.parsed_consumer_offsets,
                'zookeeper',
            )

            ordered_args, _ = mock_set_offsets.call_args
            assert ordered_args[1] == 'group1'
            assert ordered_args[2] == self.new_consumer_offsets

    def test_build_new_offsets(self, mock_kafka_client):
        new_offsets = OffsetRestore.build_new_offsets(
            mock_kafka_client,
            {'topic1': {0: 10, 1: 20}},
            {'topic1': [0, 1]},
            self.kafka_consumer_offsets,
        )

        assert new_offsets == self.new_consumer_offsets



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock
import pytest

from kafka_utils.kafka_consumer_manager. \
    commands.watermark_get import WatermarkGet


class TestGetWatermark(object):

    @pytest.yield_fixture
    def client(self):
        with mock.patch(
            'kafka_utils.kafka_consumer_manager.'
            'commands.watermark_get.KafkaToolClient',
            autospec=True,
        ) as mock_client:
            yield mock_client

    def test_get_watermark_for_topic(self, client):
        topics = '__consumer_offsets'
        client.topic_partitions = {}
        with mock.patch(
            'kafka_utils.kafka_consumer_manager.commands.'
            'watermark_get.get_watermark_for_topic',
            return_value={'test_topic': [1, 99, 3]},
            autospec=True,
        ) as mock_get_watermark:
            WatermarkGet.get_watermarks(
                client,
                topics,
                exact=True
            )
            assert mock_get_watermark.call_count == 1

    def test_get_watermark_for_regex(self, client):
        topics = '__consumer_*'
        client.topic_partitions = {}
        with mock.patch(
            'kafka_utils.kafka_consumer_manager.commands.'
            'watermark_get.get_watermark_for_regex',
            return_value={'__consumer_1': [1, 99, 3],
                          '__consumer_2': [2, 100, 2]},
            autospec=True,
        ) as mock_get_watermark:
            WatermarkGet.get_watermarks(
                client,
                topics,
                exact=False
            )
            assert mock_get_watermark.call_count == 1



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
from __future__ import print_function

import logging
import sys
from collections import defaultdict

from kafka.common import ConsumerTimeout
from kafka.common import FailedPayloadsError
from kafka.common import KafkaUnavailableError
from kafka.common import LeaderNotAvailableError
from kafka.common import NotLeaderForPartitionError
from kafka.consumer import KafkaConsumer
from kafka.util import read_short_string
from kafka.util import relative_unpack
from kazoo.exceptions import NoNodeError

from .offset_manager import OffsetManagerBase
from kafka_utils.util.offsets import get_topics_watermarks
from kafka_utils.util.zookeeper import ZK

CONSUMER_OFFSET_TOPIC = '__consumer_offsets'


class ListGroups(OffsetManagerBase):

    @classmethod
    def setup_subparser(cls, subparsers):
        parser_list_groups = subparsers.add_parser(
            "list_groups",
            description="List consumer groups.",
            add_help=False,
        )
        parser_list_groups.add_argument(
            '--storage',
            choices=['zookeeper', 'kafka', 'dual'],
            help="String describing where to fetch the committed offsets.",
            default='dual'
        )
        parser_list_groups.set_defaults(command=cls.run)

    @classmethod
    def get_zookeeper_groups(cls, cluster_config):
        '''Get the group_id of groups committed into Zookeeper.'''
        with ZK(cluster_config) as zk:
            try:
                return zk.get_children("/consumers")
            except NoNodeError:
                print(
                    "Error: No consumers node found in zookeeper",
                    file=sys.stderr,
                )

    @classmethod
    def get_kafka_groups(cls, cluster_config):
        '''Get the group_id of groups committed into Kafka.'''
        kafka_group_reader = KafkaGroupReader(cluster_config)
        try:
            return kafka_group_reader.read_groups().keys()
        except:
            print(
                "Error: No consumer offsets topic found in Kafka",
                file=sys.stderr,
            )

    @classmethod
    def print_groups(cls, groups, cluster_config):
        print("Consumer Groups:")
        for groupid in groups:
            print("\t{groupid}".format(groupid=groupid))
        print(
            "{num_groups} groups found for cluster {cluster_name} "
            "of type {cluster_type}".format(
                num_groups=len(groups),
                cluster_name=cluster_config.name,
                cluster_type=cluster_config.type,
            ),
        )

    @classmethod
    def run(cls, args, cluster_config):
        groups = set()

        if args.storage in ('dual', 'zookeeper'):
            zk_groups = cls.get_zookeeper_groups(cluster_config)
            if zk_groups:
                groups.update(zk_groups)

        if args.storage in ('dual', 'kafka'):
            kafka_groups = cls.get_kafka_groups(cluster_config)
            if kafka_groups:
                groups.update(kafka_groups)

        cls.print_groups(groups, cluster_config)


class InvalidMessageException(Exception):
    pass


class KafkaGroupReader:

    def __init__(self, kafka_config):
        self.log = logging.getLogger(__name__)
        self.kafka_config = kafka_config
        self.kafka_groups = defaultdict(set)
        self.finished_partitions = set()

    def read_groups(self):
        self.log.info("Kafka consumer running")
        self.consumer = KafkaConsumer(
            CONSUMER_OFFSET_TOPIC,
            group_id='offset_monitoring_consumer',
            bootstrap_servers=self.kafka_config.broker_list,
            auto_offset_reset='smallest',
            auto_commit_enable=False,
            consumer_timeout_ms=10000,
        )
        self.log.info("Consumer ready")
        self.watermarks = self.get_current_watermarks()
        while not self.finished():
            try:
                message = self.consumer.next()
                max_offset = self.get_max_offset(message.partition)
                if message.offset >= max_offset - 1:
                    self.finished_partitions.add(message.partition)
            except ConsumerTimeout:
                break
            except (
                    FailedPayloadsError,
                    KafkaUnavailableError,
                    LeaderNotAvailableError,
                    NotLeaderForPartitionError,
            ) as e:
                self.log.warning("Got %s, retrying", e.__class__.__name__)
            self.process_consumer_offset_message(message)
        return self.kafka_groups

    def parse_consumer_offset_message(self, message):
        key = bytearray(message.key)
        ((key_schema,), cur) = relative_unpack('>h', key, 0)
        if key_schema not in [0, 1]:
            raise InvalidMessageException()   # This is not an offset commit message
        (group, cur) = read_short_string(key, cur)
        (topic, cur) = read_short_string(key, cur)
        ((partition,), cur) = relative_unpack('>l', key, cur)
        if message.value:
            value = bytearray(message.value)
            ((value_schema,), cur) = relative_unpack('>h', value, 0)
            if value_schema not in [0, 1]:
                raise InvalidMessageException()  # Unrecognized message value
            ((offset,), cur) = relative_unpack('>q', value, cur)
        else:
            offset = None  # Offset was deleted
        return str(group), str(topic), partition, offset

    def process_consumer_offset_message(self, message):
        try:
            group, topic, partition, offset = self.parse_consumer_offset_message(message)
        except InvalidMessageException:
            return

        if offset:
            self.kafka_groups[group].add(topic)
        else:  # No offset means group deletion
            self.kafka_groups.pop(group, None)

    def get_current_watermarks(self):
        self.consumer._client.load_metadata_for_topics()
        offsets = get_topics_watermarks(
            self.consumer._client,
            [CONSUMER_OFFSET_TOPIC],
        )
        return {partition: offset for partition, offset
                in offsets[CONSUMER_OFFSET_TOPIC].iteritems()
                if offset.highmark > offset.lowmark}

    def get_max_offset(self, partition):
        return self.watermarks[partition].highmark

    def finished(self):
        return len(self.finished_partitions) >= len(self.watermarks)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from kafka_utils.kafka_check.commands.command import parse_meta_properties_file


META_PROPERTIES_CONTENT = [
    "#\n",
    "#Thu May 12 17:59:12 BST 2016\n",
    "version=0\n",
    "broker.id=123\n",
]


class TestParseMetaPropertiesFile(object):

    def test_parse_meta_properties_file(self):
        broker_id = parse_meta_properties_file(META_PROPERTIES_CONTENT)

        assert broker_id == 123

    def test_parse_meta_properties_file_empty(self):
        broker_id = parse_meta_properties_file([])

        assert broker_id is None



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from kafka_utils.kafka_check import status_code
from kafka_utils.kafka_check.status_code import terminate
from kafka_utils.util.zookeeper import ZK


class KafkaCheckCmd(object):
    """Interface used by all kafka_check commands
    The attributes cluster_config, args and zk are initialized on run().
    """

    def __init__(self):
        self.cluster_config = None
        self.args = None
        self.zk = None

    def build_subparser(self, subparsers):
        """Build the command subparser.

        :param subparsers: argpars subparsers
        :returns: subparser
        """
        raise NotImplementedError("Implement in subclass")

    def run_command(self):
        """Implement the command logic.
        When run_command is called cluster_config, args, and zk are already
        initialized.
        """
        raise NotImplementedError("Implement in subclass")

    def run(self, cluster_config, args):
        self.cluster_config = cluster_config
        self.args = args
        with ZK(self.cluster_config) as self.zk:
            if args.controller_only:
                check_run_on_controller(self.zk, self.args)
            return self.run_command()

    def add_subparser(self, subparsers):
        self.build_subparser(subparsers).set_defaults(command=self.run)


def get_controller_id(zk):
    return zk.get_json('/controller').get('brokerid')


def parse_meta_properties_file(content):
    for line in content:
        parts = line.rstrip().split("=")
        if len(parts) == 2 and parts[0] == "broker.id":
            return int(parts[1])
    return None


def read_generated_broker_id(meta_properties_path):
    """reads broker_id from meta.properties file.

    :param string meta_properties_path: path for meta.properties file
    :returns int: broker_id from meta_properties_path
    """
    try:
        with open(meta_properties_path, 'r') as f:
            broker_id = parse_meta_properties_file(f)
    except IOError:
        terminate(
            status_code.WARNING,
            "Cannot open meta.properties file: {path}".format(
                path=meta_properties_path,
            ),
        )
    except ValueError:
        terminate(status_code.WARNING, "Broker id not valid")

    if broker_id is None:
        terminate(status_code.WARNING, "Autogenerated broker id missing from data directory")

    return broker_id


def get_broker_id(data_path):
    """This function will look into the data folder to get the automatically created
    broker_id.

    :param string data_path: the path to the kafka data folder
    :returns int: the real broker_id
    """

    # Path to the meta.properties file. This is used to read the automatic broker id
    # if the given broker id is -1
    META_FILE_PATH = "{data_path}/meta.properties"

    if not data_path:
        terminate(status_code.WARNING, "You need to specify the data_path if broker_id == -1")
    meta_properties_path = META_FILE_PATH.format(data_path=data_path)
    return read_generated_broker_id(meta_properties_path)


def check_run_on_controller(zk, args):
    """Kafka 0.9 supports automatic broker ids. If args.broker_id is set to -1,
    it will call get_broker_id() for parse broker_id from meta.properties file.
    """
    if args.broker_id is None:
        terminate(status_code.WARNING, "Broker id is not specified")

    if args.broker_id != -1:
        broker_id = args.broker_id
    else:
        broker_id = get_broker_id(args.data_path)

    controller_id = get_controller_id(zk)

    # This check is only executed by the controller
    if broker_id != controller_id:
        terminate(status_code.OK, 'Broker %s is not the controller, nothing to check' % (broker_id))



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock
import pytest

from kafka_utils.kafka_consumer_manager. \
    commands.offset_rewind import OffsetRewind


class TestOffsetRewind(object):
    topics_partitions = {
        "topic1": [0, 1, 2],
        "topic2": [0, 1]
    }

    @mock.patch('kafka_utils.kafka_consumer_manager.'
                'commands.offset_rewind.KafkaToolClient')
    def test_run(self, mock_client):
        with mock.patch.object(
            OffsetRewind,
            "preprocess_args",
            spec=OffsetRewind.preprocess_args,
            return_value=self.topics_partitions,
        ), mock.patch(
            "kafka_utils.kafka_consumer_manager."
            "commands.offset_rewind.rewind_consumer_offsets",
            autospec=True
        ) as mock_rewind:
            args = mock.Mock(
                groupid="some_group",
                topic=None,
                partitions=None
            )
            cluster_config = mock.Mock()
            OffsetRewind.run(args, cluster_config)

            ordered_args, _ = mock_rewind.call_args
            assert ordered_args[1] == args.groupid
            assert ordered_args[2] == self.topics_partitions
            mock_client.return_value.load_metadata_for_topics. \
                assert_called_once_with()
            mock_client.return_value.close.assert_called_once_with()

    @mock.patch('kafka_utils.kafka_consumer_manager'
                '.commands.offset_rewind.KafkaToolClient')
    def test_run_type_error(self, mock_client):
        with mock.patch.object(
            OffsetRewind,
            "preprocess_args",
            spec=OffsetRewind.preprocess_args,
            return_value="some_string",
        ):
            args = mock.Mock(
                groupid="some_group",
                topic=None,
                partitions=None
            )
            cluster_config = mock.Mock()
            with pytest.raises(TypeError):
                OffsetRewind.run(args, cluster_config)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from kafka.common import PartitionMetadata

from kafka_utils.kafka_check.commands.under_replicated import _process_topic_partition_metadata


def test__process_topic_partition_metadata_empty():
    assert _process_topic_partition_metadata({}) == set()


def test__process_topic_partition_metadata_all_good():
    METADATA_RESPONSE = {
        'topic0': {
            0: PartitionMetadata(
                topic='topic0',
                partition=0,
                leader=13,
                replicas=(13, 100),
                isr=(13, 100),
                error=0,
            ),
            1: PartitionMetadata(
                topic='topic0',
                partition=1,
                leader=100,
                replicas=(13, 100),
                isr=(13, 100),
                error=0,
            ),
        },
        'topic1': {
            0: PartitionMetadata(
                topic='topic1',
                partition=0,
                leader=666,
                replicas=(300, 500, 666),
                isr=(300, 500, 666),
                error=0,
            ),
            1: PartitionMetadata(
                topic='topic1',
                partition=1,
                leader=300,
                replicas=(300, 500, 666),
                isr=(300, 500, 666),
                error=0,
            ),
        },
    }
    result = _process_topic_partition_metadata(METADATA_RESPONSE)
    assert result == set()


def test__process_topic_partition_metadata():
    METADATA_RESPONSE = {
        'topic0': {
            0: PartitionMetadata(
                topic='topic0',
                partition=0,
                leader=13,
                replicas=(13, 100),
                isr=(13, 100),
                error=0,
            ),
            1: PartitionMetadata(
                topic='topic0',
                partition=1,
                leader=100,
                replicas=(13, 100),
                isr=(13, 100),
                error=0,
            ),
        },
        'topic1': {
            0: PartitionMetadata(
                topic='topic1',
                partition=0,
                leader=666,
                replicas=(300, 500, 666),  # one more replica
                isr=(500, 666),
                error=9,
            ),
            1: PartitionMetadata(
                topic='topic1',
                partition=1,
                leader=300,
                replicas=(300, 500, 666),
                isr=(500, 666),
                error=0,
            ),
        },
    }

    result = _process_topic_partition_metadata(METADATA_RESPONSE)
    assert result == set([('topic1', 0)])


def test__process_topic_partition_metadata_few_for_one_broker():
    METADATA_RESPONSE = {
        'topic0': {
            0: PartitionMetadata(
                topic='topic0',
                partition=0,
                leader=100,
                replicas=(100),
                isr=(100,),
                error=9,
            ),
            1: PartitionMetadata(
                topic='topic0',
                partition=1,
                leader=100,
                replicas=(13, 100),
                isr=(13, 100),
                error=0,
            ),
        },
        'topic1': {
            0: PartitionMetadata(
                topic='topic1',
                partition=0,
                leader=666,
                replicas=(13, 500, 666),
                isr=(13, 500, 666),
                error=0,
            ),
            1: PartitionMetadata(
                topic='topic1',
                partition=1,
                leader=666,
                replicas=(666),
                isr=(666,),
                error=9,
            ),
        },
    }
    result = _process_topic_partition_metadata(METADATA_RESPONSE)

    assert result == set([('topic0', 0), ('topic1', 1)])



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock

from kafka_utils.kafka_consumer_manager. \
    commands.offset_save import OffsetSave
from kafka_utils.util.monitoring import ConsumerPartitionOffsets


class TestOffsetSave(object):
    topics_partitions = {
        "topic1": [0, 1, 2],
        "topic2": [0, 1, 2, 3],
        "topic3": [0, 1],
    }
    consumer_offsets_metadata = {
        'topic1':
        [
            ConsumerPartitionOffsets(topic='topic1', partition=0, current=10, highmark=655, lowmark=655),
            ConsumerPartitionOffsets(topic='topic1', partition=1, current=20, highmark=655, lowmark=655),
        ]
    }
    offset_data_file = {'groupid': 'group1', 'offsets': {'topic1': {0: 10, 1: 20}}}
    json_data = {'groupid': 'group1', 'offsets': {'topic1': {'0': 10, '1': 20}}}

    @mock.patch('kafka_utils.kafka_consumer_manager.'
                'commands.offset_save.KafkaToolClient')
    def test_save_offsets(self, mock_client):
        with mock.patch.object(
            OffsetSave,
            "write_offsets_to_file",
            spec=OffsetSave.write_offsets_to_file,
            return_value=[],
        ) as mock_write_offsets:
            filename = 'offset_file'
            consumer_group = 'group1'
            OffsetSave.save_offsets(
                self.consumer_offsets_metadata,
                self.topics_partitions,
                filename,
                consumer_group,
            )

            ordered_args, _ = mock_write_offsets.call_args
            assert ordered_args[0] == filename
            assert ordered_args[1] == self.offset_data_file

    @mock.patch('kafka_utils.kafka_consumer_manager.'
                'commands.offset_save.KafkaToolClient')
    def test_run(self, mock_client):
        with mock.patch.object(
            OffsetSave,
            "preprocess_args",
            spec=OffsetSave.preprocess_args,
            return_value=self.topics_partitions,
        ), mock.patch(
            "kafka_utils.kafka_consumer_manager."
            "commands.offset_save.get_consumer_offsets_metadata",
            return_value=self.consumer_offsets_metadata,
            autospec=True,
        ), mock.patch.object(
            OffsetSave,
            "write_offsets_to_file",
            spec=OffsetSave.write_offsets_to_file,
            return_value=[],
        ) as mock_write_offsets:
            args = mock.Mock(
                groupid="group1",
                json_file="some_file",
            )
            cluster_config = mock.Mock()
            OffsetSave.run(args, cluster_config)

            mock_client.return_value.load_metadata_for_topics. \
                assert_called_once_with()
            mock_client.return_value.close.assert_called_once_with()
            ordered_args, _ = mock_write_offsets.call_args
            assert ordered_args[0] == "some_file"
            assert ordered_args[1] == self.offset_data_file



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import OrderedDict

from pytest import fixture

from kafka_utils.kafka_cluster_manager.cmds.command import ClusterManagerCmd


@fixture
def orig_assignment():
    return OrderedDict([
        ((u'T1', 1), [2, 1]),
        ((u'T0', 0), [0, 1]),
        ((u'T0', 1), [1, 2]),
        ((u'T1', 0), [0, 1]),
        ((u'T2', 0), [3, 1]),
    ])


# Replica changes for (T0, 0) and (T1, 1)
# Leader only changes (T0, 1) and (T2, 0)
@fixture
def new_assignment():
    return OrderedDict([
        ((u'T0', 0), [2, 0]),
        ((u'T1', 1), [2, 3]),
        ((u'T0', 1), [2, 1]),
        ((u'T1', 0), [0, 1]),
        ((u'T2', 0), [1, 3]),
    ])


@fixture
def cmd():
    return ClusterManagerCmd()


class TestClusterManagerCmd(object):

    def test_reduced_proposed_plan_no_change(self, cmd, orig_assignment):
        # Provide same assignment
        proposed_assignment = cmd.get_reduced_assignment(
            original_assignment=orig_assignment,
            new_assignment=orig_assignment,
            max_partition_movements=1,
            max_leader_only_changes=1,
        )

        # Verify no proposed plan
        assert proposed_assignment == {}

    def test_extract_actions_unique_topics_limited_actions(self, cmd):
        movements_count = [
            ((u'T0', 0), 1),
            ((u'T0', 1), 1),
            ((u'T1', 0), 1),
            ((u'T2', 0), 1),
        ]
        red_actions = cmd._extract_actions_unique_topics(
            movements_count,
            3,
        )

        assert len(red_actions) == 3
        # Verify that all actions have unique topic
        topics = [action[0] for action in red_actions]
        assert set(topics) == set([u'T0', u'T1', u'T2'])

    def test_extract_actions_partition_movement_no_action(self, cmd):
        # In case max-allowed partition-movements is less than replication-factor
        # there is a possibility it will never converge
        movements_count = [
            ((u'T0', 0), 2),
            ((u'T0', 1), 2),
            ((u'T1', 0), 2),
            ((u'T2', 0), 2),
        ]
        red_actions = cmd._extract_actions_unique_topics(
            movements_count,
            1,
        )

        # All actions have minimum of 2 movements
        # so reduced proposed-plan is empty
        assert red_actions == []

    def test_extract_actions_partition_movements_all(self, cmd):
        movements_count = [
            ((u'T0', 0), 2),
            ((u'T0', 1), 1),
            ((u'T1', 0), 2),
            ((u'T2', 0), 1),
        ]
        red_actions = cmd._extract_actions_unique_topics(
            movements_count,
            10,
        )

        # Total movements as in proposed-plan is 6 (2+1+2+1)
        # max partition-movements allowed is 10, so all 6 partition-movements
        # allowed
        assert len(red_actions) == 4
        assert all(
            t_p in red_actions
            for t_p in ((u'T0', 0), (u'T0', 1), (u'T1', 0), (u'T2', 0))
        )

    def test_extract_actions_no_movements(self, cmd):
        movements_count = [
            ((u'T0', 0), 2),
            ((u'T0', 1), 1),
            ((u'T1', 0), 2),
            ((u'T2', 0), 1),
        ]
        red_actions = cmd._extract_actions_unique_topics(
            movements_count,
            0,
        )

        assert red_actions == []

    def test_extract_actions_unique_topics_some_actions(self, cmd):
        # Complex case
        # Total proposed-movements: 2*4 + 1 = 9
        # Max allowed movements: 5
        # Total actions in proposed_assignment: 5
        # Expected: Final assignment should have all 3 actions
        # all 3 unique topics
        movements_count = [
            ((u'T0', 0), 2),
            ((u'T1', 1), 2),
            ((u'T0', 1), 2),
            ((u'T1', 0), 2),
            ((u'T2', 0), 1),
        ]
        red_actions = cmd._extract_actions_unique_topics(
            movements_count,
            5,
        )

        assert len(red_actions) == 3
        # Verify T0, T1 and T2 are all in the result
        topics = [action[0] for action in red_actions]
        assert set(topics) == set([u'T0', u'T1', u'T2'])

    def test_reduced_proposed_plan_zero_changes(
        self,
        cmd,
        orig_assignment,
        new_assignment,
    ):
        # Provide less than max_changes parameter
        proposed_assignment = cmd.get_reduced_assignment(
            orig_assignment,
            new_assignment,
            max_partition_movements=0,
            max_leader_only_changes=0,
        )

        # Verify no proposed plan
        assert proposed_assignment == {}

    def test_reduced_proposed_plan_zero_negative_changes(
        self,
        cmd,
        orig_assignment,
        new_assignment,
    ):
        proposed_assignment = cmd.get_reduced_assignment(
            orig_assignment,
            new_assignment,
            max_partition_movements=-1,
            max_leader_only_changes=-1,
        )

        # Verify no proposed plan
        assert proposed_assignment == {}

    def test_reduced_proposed_plan_empty_new_assignment(self, cmd, orig_assignment):
        # Provide empty assignment
        proposed_assignment = cmd.get_reduced_assignment(
            orig_assignment,
            new_assignment={},
            max_partition_movements=1,
            max_leader_only_changes=1,
        )

        # Verify no proposed plan
        assert proposed_assignment == {}

    def test_reduced_proposed_plan_empty_original_assignment(self, cmd, new_assignment):
        proposed_assignment = cmd.get_reduced_assignment(
            original_assignment={},
            new_assignment=new_assignment,
            max_partition_movements=1,
            max_leader_only_changes=1,
        )

        # Verify no proposed plan
        assert proposed_assignment == {}

    def test_reduced_proposed_plan_only_partitions(
        self,
        cmd,
        orig_assignment,
        new_assignment,
    ):
        result = cmd.get_reduced_assignment(
            orig_assignment,
            new_assignment,
            max_partition_movements=2,
            max_leader_only_changes=0,
        )

        assert len(result) == 2
        # T2 not in result because leader only change
        assert (u'T2', 0) not in result
        # T1 no changes for 0
        assert (u'T1', 0) not in result and (u'T1', 1) in result
        # T0 leader only changes for 1
        assert (u'T0', 0) in result and (u'T0', 1) not in result

    def test_reduced_proposed_plan_only_leaders(
        self,
        cmd,
        orig_assignment,
        new_assignment,
    ):
        result = cmd.get_reduced_assignment(
            orig_assignment,
            new_assignment,
            max_partition_movements=0,
            max_leader_only_changes=2,
        )

        assert len(result) == 2
        # T2 leader only change for 0
        assert (u'T2', 0) in result
        # T1 no leader only changes
        assert (u'T1', 0) not in result and (u'T1', 1) not in result
        # T0 leader only changes for 1
        assert (u'T0', 0) not in result and (u'T0', 1) in result

    def test_reduced_proposed_plan(
        self,
        cmd,
        orig_assignment,
        new_assignment,
    ):
        result = cmd.get_reduced_assignment(
            orig_assignment,
            new_assignment,
            max_partition_movements=2,
            max_leader_only_changes=2,
        )

        assert len(result) == 4
        assert (u'T2', 0) in result
        # T1 no changes for 0
        assert (u'T1', 0) not in result and (u'T1', 1) in result
        assert (u'T0', 0) in result and (u'T0', 1) in result



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging

from .command import ClusterManagerCmd
from kafka_utils.kafka_cluster_manager.cluster_info.stats import imbalance_value_all
from kafka_utils.util.validation import plan_to_assignment


class StatsCmd(ClusterManagerCmd):

    def __init__(self):
        super(StatsCmd, self).__init__()
        self.log = logging.getLogger(self.__class__.__name__)

    def build_subparser(self, subparsers):
        subparser = subparsers.add_parser(
            'stats',
            description='Show imbalance statistics of cluster topology',
            help='This command is used to display imbalance statistics of current '
            'cluster-topology or cluster-topology after given assignment is '
            'applied.',
        )
        subparser.add_argument(
            '--read-from-file',
            dest='plan_file_path',
            metavar='<reassignment-plan-file-path>',
            type=str,
            help='Read the partition assignment from json file. Example format:'
            ' {"version": 1, "partitions": [{"topic": "foo", "partition": 1, '
            '"replicas": [1,2,3]}]}',
        )
        return subparser

    def run_command(self, cluster_topology):
        if self.args.plan_file_path:
            base_assignment = cluster_topology.assignment

            print('Integrating given assignment-plan in current cluster-topology.')
            cluster_topology.update_cluster_topology(self.get_assignment())
            self.imbalance_stats(cluster_topology, base_assignment)
        else:
            self.imbalance_stats(cluster_topology)

    def log_imbalance_stats(self, imbal):
        net_imbalance = (
            imbal['replica_cnt'] +
            imbal['net_part_cnt_per_rg'] +
            imbal['topic_partition_cnt']
        )
        print(
            'Replication-group imbalance (replica-count): {imbal_repl}\n'
            'Net Partition-count imbalance/replication-group: '
            '{imbal_part_rg}\nNet Partition-count imbalance: {imbal_part}\n'
            'Topic-partition-count imbalance: {imbal_tp}\n'
            'Net-cluster imbalance (excluding leader-imbalance): '
            '{imbal_net}'.format(
                imbal_part_rg=imbal['net_part_cnt_per_rg'],
                imbal_repl=imbal['replica_cnt'],
                imbal_part=imbal['partition_cnt'],
                imbal_tp=imbal['topic_partition_cnt'],
                imbal_net=net_imbalance,
            )
        )
        net_imbalance_with_leaders = net_imbalance + imbal['leader_cnt']
        print(
            'Leader-count imbalance: {imbal_leader}\n'
            'Net-cluster imbalance (including leader-imbalance): '
            '{imbal}'.format(
                imbal=net_imbalance_with_leaders,
                imbal_leader=imbal['leader_cnt'],
            )
        )

        print(
            'Total partition-movements: {total_movements}'.format(
                total_movements=imbal['total_movements'],
            ),
        )

    def imbalance_stats(self, ct, base_assignment=None):
        print('Calculating cluster imbalance statistics...')
        initial_imbal = imbalance_value_all(ct, base_assignment)
        self.log_imbalance_stats(initial_imbal)
        total_imbal = (
            initial_imbal['replica_cnt'] +
            initial_imbal['net_part_cnt_per_rg'] +
            initial_imbal['topic_partition_cnt'] +
            initial_imbal['partition_cnt'] +
            initial_imbal['leader_cnt']
        )
        if total_imbal == 0:
            print('Cluster is currently balanced!')

    def get_assignment(self):
        """Parse the given json plan in dict format."""
        try:
            plan = json.loads(open(self.args.plan_file_path).read())
            return plan_to_assignment(plan)
        except IOError:
            self.log.exception(
                'Given json file {file} not found.'
                .format(file=self.args.plan_file_path),
            )
            raise
        except ValueError:
            self.log.exception(
                'Given json file {file} could not be decoded.'
                .format(file=self.args.plan_file_path),
            )
            raise
        except KeyError:
            self.log.exception(
                'Given json file {file} could not be parsed in desired format.'
                .format(file=self.args.plan_file_path),
            )
            raise



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock
import pytest
from kazoo.exceptions import NoNodeError
from kazoo.exceptions import ZookeeperError

from kafka_utils.kafka_consumer_manager.commands. \
    unsubscribe_topics import KafkaUnsubscriber
from kafka_utils.kafka_consumer_manager.commands. \
    unsubscribe_topics import UnsubscribeTopics
from kafka_utils.kafka_consumer_manager.commands. \
    unsubscribe_topics import ZookeeperUnsubscriber


class TestUnsubscribeTopics(object):

    @pytest.yield_fixture
    def client(self):
        with mock.patch(
            'kafka_utils.kafka_consumer_manager.'
            'commands.unsubscribe_topics.KafkaToolClient',
            autospec=True,
        ) as mock_client:
            yield mock_client

    @pytest.yield_fixture
    def zk(self):
        with mock.patch(
            'kafka_utils.kafka_consumer_manager.'
            'commands.unsubscribe_topics.ZK',
            autospec=True
        ) as mock_zk:
            yield mock_zk

    @pytest.yield_fixture
    def zk_unsubscriber(self):
        with mock.patch(
            'kafka_utils.kafka_consumer_manager.'
            'commands.unsubscribe_topics.ZookeeperUnsubscriber',
            autospec=True,
        ) as mock_zk_unsub:
            yield mock_zk_unsub

    @pytest.yield_fixture
    def kafka_unsubscriber(self):
        with mock.patch(
            'kafka_utils.kafka_consumer_manager.'
            'commands.unsubscribe_topics.KafkaUnsubscriber',
            autospec=True,
        ) as mock_kafka_unsub:
            yield mock_kafka_unsub

    topics_partitions = {
        'topic1': [0, 1, 2],
        'topic2': [0, 1],
    }

    cluster_config = mock.Mock(zookeeper='some_ip')

    def test_run_unsubscribe_zk(self, client, zk, zk_unsubscriber, kafka_unsubscriber):
        with mock.patch.object(
                UnsubscribeTopics,
                'preprocess_args',
                spec=UnsubscribeTopics.preprocess_args,
                return_value=self.topics_partitions,
        ):
            args = mock.Mock(storage='zookeeper')

            UnsubscribeTopics.run(args, self.cluster_config)

            zk_obj = zk_unsubscriber.return_value
            kafka_obj = kafka_unsubscriber.return_value

            assert zk_obj.unsubscribe_topic.call_count == 1
            assert kafka_obj.unsubscribe_topic.call_count == 0

    def test_run_unsubscribe_kafka(self, client, zk, zk_unsubscriber, kafka_unsubscriber):
        with mock.patch.object(
                UnsubscribeTopics,
                'preprocess_args',
                spec=UnsubscribeTopics.preprocess_args,
                return_value=self.topics_partitions,
        ):
            args = mock.Mock(storage='kafka')

            UnsubscribeTopics.run(args, self.cluster_config)

            zk_obj = zk_unsubscriber.return_value
            kafka_obj = kafka_unsubscriber.return_value

            assert zk_obj.unsubscribe_topic.call_count == 0
            assert kafka_obj.unsubscribe_topic.call_count == 1

    def test_unsubscribe_some_partitions_left(self, zk):
        zk_obj = zk.return_value
        zk_obj.get_my_subscribed_partitions.return_value = [3]

        unsubscriber = ZookeeperUnsubscriber(zk_obj)
        unsubscriber.unsubscribe_topic(
            'some_group',
            'topic1',
            [0, 1, 2],
            self.topics_partitions,
        )

        calls = [
            mock.call(
                'some_group',
                'topic1',
                [0, 1, 2],
            ),
        ]

        assert zk_obj.delete_topic_partitions.call_args_list == calls
        # Delete topic should not be called because the group is still
        # subscribed to some topic partitions
        assert not zk_obj.delete_topic.called

    def test_unsubscribe_wipe_all_partitions(self, zk):
        zk_obj = zk.return_value
        zk_obj.get_my_subscribed_partitions.return_value = []

        unsubscriber = ZookeeperUnsubscriber(zk_obj)
        unsubscriber.unsubscribe_topic(
            'some_group',
            'topic1',
            [0, 1, 2],
            self.topics_partitions,
        )

        calls = [
            mock.call(
                'some_group',
                'topic1',
                [0, 1, 2],
            ),
        ]

        assert zk_obj.delete_topic_partitions.call_args_list == calls
        assert zk_obj.delete_topic.call_args_list == [
            mock.call('some_group', 'topic1'),
        ]

    def test_unsubscribe_wipe_default_partitions(self, zk):
        zk_obj = zk.return_value
        zk_obj.get_my_subscribed_partitions.return_value = []

        unsubscriber = ZookeeperUnsubscriber(zk_obj)
        unsubscriber.unsubscribe_topic(
            'some_group',
            'topic1',
            None,
            self.topics_partitions,
        )

        assert zk_obj.delete_topic_partitions.call_count == 0
        assert zk_obj.delete_topic.call_args_list == [
            mock.call('some_group', 'topic1'),
        ]

    def test_unsubscribe_wipe_default_topics(self, zk):
        zk_obj = zk.return_value
        zk_obj.get_my_subscribed_partitions.return_value = []

        unsubscriber = ZookeeperUnsubscriber(zk_obj)
        unsubscriber.unsubscribe_topic(
            'some_group',
            None,
            None,
            self.topics_partitions,
        )

        assert zk_obj.delete_topic_partitions.call_count == 0
        assert sorted(zk_obj.delete_topic.call_args_list) == sorted(
            [
                mock.call('some_group', 'topic1'),
                mock.call('some_group', 'topic2'),
            ],
        )

    def test_unsubscribe_no_node_error(self, zk):
        zk_obj = zk.return_value
        zk_obj.delete_topic_partitions.side_effect = NoNodeError("Boom!")

        unsubscriber = ZookeeperUnsubscriber(zk_obj)
        unsubscriber.unsubscribe_topic(
            'some_group',
            'topic1',
            [0, 1, 2],
            self.topics_partitions,
        )

        assert zk_obj.delete_topic_partitions.called

    def test_unsubscribe_any_other_exception(self, zk):
        zk_obj = zk.return_value
        zk_obj.delete_topic_partitions.side_effect = ZookeeperError("Boom!")

        unsubscriber = ZookeeperUnsubscriber(zk_obj)
        with pytest.raises(ZookeeperError):
            unsubscriber.unsubscribe_topic(
                'some_group',
                'topic1',
                [0, 1, 2],
                self.topics_partitions,
            )

    def test_unsubscribe_topic_kafka_storage(self, client):
        offsets = {'topic1': {0: 100}}
        new_offsets = {'topic1': {0: -1}}

        with mock.patch(
            'kafka_utils.kafka_consumer_manager.'
            'commands.unsubscribe_topics.get_current_consumer_offsets',
            autospec=True,
            return_value=offsets,
        ) as mock_get, mock.patch(
            'kafka_utils.kafka_consumer_manager.'
            'commands.unsubscribe_topics.set_consumer_offsets',
            autospec=True,
        ) as mock_set:
            unsubscriber = KafkaUnsubscriber(client)
            unsubscriber.delete_topic('some_group', 'topic1')

            assert mock_get.call_count == 1
            assert mock_set.call_count == 1
            assert mock_set.call_args_list == [
                mock.call(
                    client,
                    'some_group',
                    new_offsets,
                    offset_storage='kafka',
                ),
            ]



from __future__ import print_function

import mock

from kafka_utils.kafka_corruption_check import main


FILE_LINE = "Dumping /path/to/file"
CORRECT_LINE = "offset: 2413 position: 173 isvalid: true payloadsize: 3 magic: 0 compresscodec: NoCompressionCodec crc: 1879665293"
INVALID_LINE = "offset: 2413 position: 173 isvalid: false payloadsize: 3 magic: 0 compresscodec: NoCompressionCodec crc: 1879665293"


def mock_output(values):
    output = mock.Mock()
    output.readlines.return_value = values
    return output


def test_find_files_cmd_min():
    cmd = main.find_files_cmd("path", 60, None, None)
    assert cmd == 'find "path" -type f -name "*.log" -mmin -60'


def test_find_files_cmd_start():
    cmd = main.find_files_cmd("path", None, "START", None)
    assert cmd == 'find "path" -type f -name "*.log" -newermt "START"'


def test_find_files_cmd_range():
    cmd = main.find_files_cmd("path", None, "START", "END")
    assert cmd == 'find "path" -type f -name "*.log" -newermt "START" \! -newermt "END"'


@mock.patch(
    "kafka_utils.kafka_corruption_check."
    "main.print_line",
)
def test_parse_output_correct(mock_print):
    main.parse_output("HOST", mock_output([CORRECT_LINE]))
    assert not mock_print.called


@mock.patch(
    "kafka_utils.kafka_corruption_check."
    "main.print_line",
)
def test_parse_output_invalid(mock_print):
    main.parse_output("HOST", mock_output([FILE_LINE, INVALID_LINE]))
    mock_print.assert_called_once_with("HOST", "/path/to/file", INVALID_LINE, "ERROR")


@mock.patch(
    "kafka_utils.kafka_corruption_check."
    "main.get_partition_leaders",
    return_value={"t0-0": 1, "t0-1": 2},
)
def test_filter_leader_files(mock_get_partition):
    filtered = main.filter_leader_files(None, [(1, "host1", ["a/kafka-logs/t0-0/0123.log",
                                                             "a/kafka-logs/t2-0/0123.log",
                                                             "a/kafka-logs/t0-1/0123.log"]),
                                               (2, "host2", ["a/kafka-logs/t0-0/0123.log",
                                                             "a/kafka-logs/t0-1/0123.log"])])
    assert filtered == [(1, 'host1', ['a/kafka-logs/t0-0/0123.log',
                                      'a/kafka-logs/t2-0/0123.log']),
                        (2, 'host2', ['a/kafka-logs/t0-1/0123.log'])]



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock
import pytest
from kazoo.exceptions import ZookeeperError

from kafka_utils.kafka_consumer_manager.commands. \
    delete_group import DeleteGroup


class TestDeleteGroup(object):

    @pytest.yield_fixture
    def client(self):
        with mock.patch(
                'kafka_utils.kafka_consumer_manager.'
                'commands.delete_group.KafkaToolClient',
                autospec=True,
        ) as mock_client:
            yield mock_client

    @pytest.yield_fixture
    def zk(self):
        with mock.patch(
                'kafka_utils.kafka_consumer_manager.'
                'commands.delete_group.ZK',
                autospec=True
        ) as mock_zk:
            mock_zk.return_value.__enter__.return_value = mock_zk.return_value
            yield mock_zk

    @pytest.yield_fixture
    def offsets(self):
        yield {'topic1': {0: 100}}

    def test_run_wipe_delete_group(self, client, zk, offsets):
        with mock.patch.object(
                DeleteGroup,
                'preprocess_args',
                spec=DeleteGroup.preprocess_args,
                return_value=offsets,
        ):
            args = mock.Mock(
                groupid="some_group",
                storage="zookeeper",
            )
            cluster_config = mock.Mock(zookeeper='some_ip')

            DeleteGroup.run(args, cluster_config)

            obj = zk.return_value
            assert obj.delete_group.call_args_list == [
                mock.call(args.groupid),
            ]

    def test_run_wipe_delete_group_error(self, client, zk, offsets):
        with mock.patch.object(
                DeleteGroup,
                'preprocess_args',
                spec=DeleteGroup.preprocess_args,
                return_value=offsets,
        ):
            obj = zk.return_value.__enter__.return_value
            obj.__exit__.return_value = False
            obj.delete_group.side_effect = ZookeeperError("Boom!")
            args = mock.Mock(
                groupid="some_group",
                storage="zookeeper",
            )
            cluster_config = mock.Mock(zookeeper='some_ip')

            with pytest.raises(ZookeeperError):
                DeleteGroup.run(args, cluster_config)

    def test_delete_topic_kafka_storage(self, client, offsets):
        new_offsets = {'topic1': {0: -1}}

        with mock.patch(
            'kafka_utils.kafka_consumer_manager.'
            'commands.delete_group.set_consumer_offsets',
            autospec=True,
        ) as mock_set:
            DeleteGroup.delete_group_kafka(client, 'some_group', offsets)

            assert mock_set.call_count == 1
            assert mock_set.call_args_list == [
                mock.call(
                    client,
                    'some_group',
                    new_offsets,
                    offset_storage='kafka',
                ),
            ]



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This files contains supporting api's required to evaluate stats of the
cluster at any given time.
"""
from collections import defaultdict
from math import sqrt

from .display import display_leader_count_per_broker
from .display import display_partition_count_per_broker
from .display import display_same_replica_count_rg
from .display import display_same_topic_partition_count_broker
from .util import compute_optimum
from .util import get_leaders_per_broker
from .util import get_partitions_per_broker


# Get imbalance stats
def standard_deviation(data):
    """Return standard deviation for given data with list of values."""
    avg_data = sum(data) / len(data)
    variance = map(lambda x: (x - avg_data) ** 2, data)
    avg_variance = sum(variance) / len(data)
    return sqrt(avg_variance)


def get_net_imbalance(count_per_broker):
    """Calculate and return net imbalance based on given count of
    partitions or leaders per broker.

    Net-imbalance in case of partitions implies total number of
    extra partitions from optimal count over all brokers.
    This is also implies, the minimum number of partition movements
    required for overall balancing.

    For leaders, net imbalance implies total number of extra brokers
    as leaders from optimal count.
    """
    net_imbalance = 0
    opt_count, extra_allowed = \
        compute_optimum(len(count_per_broker), sum(count_per_broker))
    for count in count_per_broker:
        extra_cnt, extra_allowed = \
            get_extra_element_count(count, opt_count, extra_allowed)
        net_imbalance += extra_cnt
    return net_imbalance


def get_extra_element_count(curr_count, opt_count, extra_allowed_cnt):
    """Evaluate and return extra same element count based on given values.

    :key-term:
    group:  In here group can be any base where elements are place
            i.e. replication-group while placing replicas (elements)
            or  brokers while placing partitions (elements).
    element:  Generic term for units which are optimally placed over group.

    :params:
    curr_count: Given count
    opt_count:  Optimal count for each group.
    extra_allowed_cnt:  Count of groups which can have 1 extra element
                   _    on each group.
    """
    if curr_count > opt_count:
        # We still can allow 1 extra count
        if extra_allowed_cnt > 0:
            extra_allowed_cnt -= 1
            extra_cnt = curr_count - opt_count - 1
        else:
            extra_cnt = curr_count - opt_count
    else:
        extra_cnt = 0
    return extra_cnt, extra_allowed_cnt


# Get imbalance stats
def get_replication_group_imbalance_stats(rgs, partitions):
    """Calculate extra replica count replica count over each replication-group
    and net extra-same-replica count.
    """
    tot_rgs = len(rgs)
    extra_replica_cnt_per_rg = defaultdict(int)
    for partition in partitions:
        # Get optimal replica-count for each partition
        opt_replica_cnt, extra_replicas_allowed = \
            compute_optimum(tot_rgs, partition.replication_factor)

        # Extra replica count for each rg
        for rg in rgs:
            replica_cnt_rg = rg.count_replica(partition)
            extra_replica_cnt, extra_replicas_allowed = \
                get_extra_element_count(
                    replica_cnt_rg,
                    opt_replica_cnt,
                    extra_replicas_allowed,
                )
            extra_replica_cnt_per_rg[rg.id] += extra_replica_cnt

    # Evaluate net imbalance across all replication-groups
    net_imbalance = sum(extra_replica_cnt_per_rg.values())
    return net_imbalance, extra_replica_cnt_per_rg


def get_leader_imbalance_stats(brokers):
    """Return for each broker the number of times it is assigned as preferred
    leader. Also return net leader-imbalance and imbalance stdev.
    """
    leaders_per_broker = get_leaders_per_broker(brokers)

    # Calculate standard deviation of leader imbalance
    stdev_imbalance = standard_deviation(leaders_per_broker.values())

    # Calculation net imbalance
    net_imbalance = get_net_imbalance(leaders_per_broker.values())
    leaders_per_broker_id = dict(
        (broker.id, count)
        for broker, count in leaders_per_broker.iteritems()
    )
    return stdev_imbalance, net_imbalance, leaders_per_broker_id


def get_topic_imbalance_stats(brokers, topics):
    """Return count of topics and partitions on each broker having multiple
    partitions of same topic.

    :rtype dict(broker_id: same-topic-partition count)
    Example:
    Total-brokers (b1, b2): 2
    Total-partitions of topic t1: 5
    (b1 has 4 partitions), (b2 has 1 partition)
    opt-count: 5/2 = 2
    extra-count: 5%2 = 1
    i.e. 1 broker can have 2 + 1 = 3 partitions
    and rest of brokers can have 2 partitions for given topic
    Extra-partition or imbalance:
    b1: current-partitions - optimal-count = 4 - 2 - 1(extra allowed) = 1
    Net-imbalance = 1
    """
    extra_partition_cnt_per_broker = defaultdict(int)
    tot_brokers = len(brokers)
    for topic in topics:
        # Optimal partition-count per topic per broker
        total_partition_replicas = \
            len(topic.partitions) * topic.replication_factor
        opt_partition_cnt, extra_partitions_allowed = \
            compute_optimum(tot_brokers, total_partition_replicas)
        # Get extra-partition count per broker for each topic
        for broker in brokers:
            partition_cnt_broker = broker.count_partitions(topic)
            extra_partitions, extra_partitions_allowed = \
                get_extra_element_count(
                    partition_cnt_broker,
                    opt_partition_cnt,
                    extra_partitions_allowed,
                )
            extra_partition_cnt_per_broker[broker.id] += extra_partitions

    # Net extra partitions over all brokers
    net_imbalance = sum(extra_partition_cnt_per_broker.itervalues())
    return net_imbalance, extra_partition_cnt_per_broker


def get_partition_imbalance_stats(brokers):
    """Return partition count for each broker and net imbalance due to extra
    partition count on each broker.
    """
    partitions_per_broker = get_partitions_per_broker(brokers)

    # Calculate standard deviation of partition imbalance
    stdev_imbalance = standard_deviation(partitions_per_broker.values())

    # Net total imbalance of partition count over all brokers
    net_imbalance = get_net_imbalance(partitions_per_broker.values())
    partitions_per_broker_id = dict(
        (partition.id, count)
        for partition, count in partitions_per_broker.iteritems()
    )
    return stdev_imbalance, net_imbalance, partitions_per_broker_id


def calculate_partition_movement(prev_assignment, curr_assignment):
    """Calculate the partition movements from initial to current assignment.
    Algorithm:
        For each partition in initial assignment
            # If replica set different in current assignment:
                # Get Difference in sets
    :rtype: tuple
    dict((partition,  (from_broker_set, to_broker_set)), total_movements
    """
    total_movements = 0
    movements = {}
    for prev_partition, prev_replicas in prev_assignment.iteritems():
        curr_replicas = curr_assignment[prev_partition]
        diff = len(set(curr_replicas) - set(prev_replicas))
        if diff:
            total_movements += diff
            movements[prev_partition] = (
                (set(prev_replicas) - set(curr_replicas)),
                (set(curr_replicas) - set(prev_replicas)),
            )
    return movements, total_movements


def partition_imbalance(rgs, brokers, cluster_wide=False):
    """Report partition count imbalance over brokers in current cluster
    state for each replication-group.
    """
    # Get broker-imbalance stats cluster-wide or for each replication-group
    if cluster_wide:
        return [(get_partition_imbalance_stats(brokers))]
    else:
        return [(get_partition_imbalance_stats(rg.brokers)) for rg in rgs]


def leader_imbalance(brokers):
    """Report leader imbalance count over each broker."""
    return get_leader_imbalance_stats(brokers)


def topic_imbalance(brokers, topics):
    """Return count of topics and partitions on each broker having multiple
    partitions of same topic.
    """
    return get_topic_imbalance_stats(brokers, topics)


def replication_group_imbalance(rgs, partitions):
    """Calculate same replica count over each replication-group."""
    return get_replication_group_imbalance_stats(rgs, partitions)


def imbalance_value_all(ct, base_assignment=None):
    """Evaluate imbalance statistics based on given params.

    Params represent the layer for which imbalance needs to be calculated
    """
    # Partition-count imbalance
    (stdev_imbalance, imbal_part, partitions_per_broker) = \
        partition_imbalance(
            ct.rgs.values(),
            ct.brokers.values(),
            cluster_wide=True,
    )[0]
    imbal_part_per_rg = partition_imbalance(
        ct.rgs.values(),
        ct.brokers.values(),
    )
    net_imbal_part_per_rg = sum(imbal_part_rg[1] for imbal_part_rg in imbal_part_per_rg)

    # Duplicate-replica-count imbalance
    imbal_replica, duplicate_replica_count_per_rg = \
        replication_group_imbalance(ct.rgs.values(), ct.partitions.values())

    # Same topic-partition count
    imbal_topic, same_topic_partition_count_per_broker = \
        topic_imbalance(ct.brokers.values(), ct.topics.values())

    imbal_leader = 0
    # Leader-count imbalance
    stdev_imbalance, imbal_leader, leaders_per_broker = \
        leader_imbalance(ct.brokers.values())

    # Display stats to stdout
    display_partition_count_per_broker(
        partitions_per_broker,
        stdev_imbalance,
        imbal_part,
    )
    display_same_replica_count_rg(duplicate_replica_count_per_rg, imbal_replica)
    display_same_topic_partition_count_broker(
        same_topic_partition_count_per_broker,
        imbal_topic,
    )
    display_leader_count_per_broker(
        leaders_per_broker,
        stdev_imbalance,
        imbal_leader,
    )

    if base_assignment:
        total_movements = calculate_partition_movement(
            base_assignment,
            ct.assignment,
        )[1]
    else:
        total_movements = 0
    return {
        'net_part_cnt_per_rg': net_imbal_part_per_rg,
        'partition_cnt': imbal_part,
        'replica_cnt': imbal_replica,
        'topic_partition_cnt': imbal_topic,
        'leader_cnt': imbal_leader,
        'total_movements': total_movements,
    }



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib
import sys

import mock
from kazoo.exceptions import NoNodeError

from kafka_utils.kafka_consumer_manager. \
    commands.list_groups import ListGroups


class TestListGroups(object):

    @contextlib.contextmanager
    def mock_kafka_info(self):
        with mock.patch(
            "kafka_utils.kafka_consumer_manager."
            "commands.list_groups.ZK",
            autospec=True
        ) as mock_ZK, mock.patch(
            "kafka_utils.kafka_consumer_manager."
            "commands.list_groups.KafkaGroupReader",
            autospec=True
        ) as mock_kafka_reader:
            mock_ZK.return_value.__enter__.return_value = mock_ZK
            yield mock_ZK, mock_kafka_reader

    def test_get_zookeeper_groups(self):
        with self.mock_kafka_info() as (mock_ZK, _):
            expected_groups = ['group1', 'group2', 'group3']

            obj = mock_ZK.return_value.__enter__.return_value
            obj.get_children.return_value = expected_groups

            cluster_config = mock.Mock(zookeeper='some_ip', type='some_cluster_type')
            cluster_config.configure_mock(name='some_cluster_name')

            assert ListGroups.get_zookeeper_groups(cluster_config) == expected_groups
            assert obj.get_children.call_count == 1

    @mock.patch("kafka_utils.kafka_consumer_manager.commands.list_groups.print", create=True)
    def test_get_zookeeper_groups_error(self, mock_print):
        with self.mock_kafka_info() as (mock_ZK, _):
            obj = mock_ZK.return_value.__enter__.return_value
            obj.__exit__.return_value = False
            cluster_config = mock.Mock(zookeeper='some_ip')
            obj.get_children.side_effect = NoNodeError("Boom!")

            ListGroups.get_zookeeper_groups(cluster_config)
            mock_print.assert_any_call(
                "Error: No consumers node found in zookeeper",
                file=sys.stderr,
            )

    def test_get_kafka_groups(self):
        with self.mock_kafka_info() as (_, mock_kafka_reader):
            expected_groups = {
                'group_name': ['topic1', 'topic2'],
                'group2': ['topic3', 'topic4']
            }

            m = mock_kafka_reader.return_value
            m.read_groups.return_value = expected_groups

            cluster_config = mock.Mock(zookeeper='some_ip', type='some_cluster_type')
            cluster_config.configure_mock(name='some_cluster_name')

            result = ListGroups.get_kafka_groups(cluster_config)
            assert result == expected_groups.keys()
            assert m.read_groups.call_count == 1

    @mock.patch("kafka_utils.kafka_consumer_manager.commands.list_groups.print", create=True)
    def test_get_kafka_groups_error(self, mock_print):
        with self.mock_kafka_info() as (_, mock_kafka_reader):
            m = mock_kafka_reader.return_value
            m.read_groups.side_effect = Exception("Boom!")

            cluster_config = mock.Mock(zookeeper='some_ip', type='some_cluster_type')
            cluster_config.configure_mock(name='some_cluster_name')

            ListGroups.get_kafka_groups(cluster_config)
            mock_print.assert_any_call(
                "Error: No consumer offsets topic found in Kafka",
                file=sys.stderr,
            )

    @mock.patch("kafka_utils.kafka_consumer_manager.commands.list_groups.print", create=True)
    def test_print_groups(self, mock_print):
        groups = ['group1', 'group2', 'group3']

        cluster_config = mock.Mock(zookeeper='some_ip', type='some_cluster_type')
        cluster_config.configure_mock(name='some_cluster_name')

        expected_print = [
            mock.call("Consumer Groups:"),
            mock.call("\tgroup1"),
            mock.call("\tgroup2"),
            mock.call("\tgroup3"),
            mock.call("3 groups found for cluster some_cluster_name "
                      "of type some_cluster_type"),
        ]

        ListGroups.print_groups(groups, cluster_config)
        assert mock_print.call_args_list == expected_print



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import logging
from collections import OrderedDict

from .broker import Broker
from .error import BrokerDecommissionError
from .error import EmptyReplicationGroupError
from .error import InvalidBrokerIdError
from .error import InvalidPartitionError
from .error import NotEligibleGroupError
from .error import RebalanceError
from .partition import Partition
from .rg import ReplicationGroup
from .topic import Topic
from .util import compute_optimum
from .util import separate_groups


class ClusterTopology(object):
    """Represent a Kafka cluster and functionalities supported over the cluster.

        :param assignment: cluster assignment is a dict (topic, partition): replicas
        :param brokers: dict representing the active brokers of the
            cluster broker_id: metadata (metadata is the content of the zookeeper
            node of the broker)
        :param extract_group: function used to extract the replication group
            from each broker. The extract_group function is called for each
            broker passing the Broker object as argument. It should return a
            string representing the ReplicationGroup id.
    """

    def __init__(self, assignment, brokers, extract_group=lambda x: None):
        self.extract_group = extract_group
        self.log = logging.getLogger(self.__class__.__name__)
        self.topics = {}
        self.rgs = {}
        self.brokers = {}
        self.partitions = {}
        self._build_brokers(brokers)
        self._build_partitions(assignment)
        self.log.debug(
            'Total partitions in cluster {partitions}'.format(
                partitions=len(self.partitions),
            ),
        )
        self.log.debug(
            'Total replication-groups in cluster {rgs}'.format(
                rgs=len(self.rgs),
            ),
        )
        self.log.debug(
            'Total brokers in cluster {brokers}'.format(
                brokers=len(self.brokers),
            ),
        )

    def _build_brokers(self, brokers):
        """Build broker objects using broker-ids."""
        for broker_id, metadata in brokers.iteritems():
            self.brokers[broker_id] = self._create_broker(broker_id, metadata)

    def _create_broker(self, broker_id, metadata=None):
        """Create a broker object and assign to a replication group.
        A broker object with no metadata is considered inactive.
        An inactive broker may or may not belong to a group.
        """
        broker = Broker(broker_id, metadata)
        if not metadata:
            broker.mark_inactive()
        rg_id = self.extract_group(broker)
        group = self.rgs.setdefault(rg_id, ReplicationGroup(rg_id))
        group.add_broker(broker)
        broker.replication_group = group
        return broker

    def _build_partitions(self, assignment):
        """Builds all partition objects and update corresponding broker and
        topic objects.
        """
        self.partitions = {}
        for partition_name, replica_ids in assignment.iteritems():
            # Get topic
            topic_id = partition_name[0]
            partition_id = partition_name[1]
            topic = self.topics.setdefault(
                topic_id,
                Topic(topic_id, replication_factor=len(replica_ids))
            )

            # Creating partition object
            partition = Partition(topic, partition_id)
            self.partitions[partition_name] = partition
            topic.add_partition(partition)

            # Updating corresponding broker objects
            for broker_id in replica_ids:
                # Check if broker-id is present in current active brokers
                if broker_id not in self.brokers.keys():
                    self.log.warning(
                        "Broker %s containing partition %s is not in "
                        "active brokers.",
                        broker_id,
                        partition,
                    )
                    self.brokers[broker_id] = self._create_broker(broker_id)

                self.brokers[broker_id].add_partition(partition)

    @property
    def assignment(self):
        assignment = {}
        for partition in self.partitions.itervalues():
            assignment[
                (partition.topic.id, partition.partition_id)
            ] = [broker.id for broker in partition.replicas]
        # assignment map created in sorted order for deterministic solution
        return OrderedDict(sorted(assignment.items(), key=lambda t: t[0]))

    def rebalance_replication_groups(self):
        """Rebalance partitions over replication groups.

        First step involves rebalancing replica-count for each partition across
        replication-groups.
        Second step involves rebalancing partition-count across replication-groups
        of the cluster.
        """
        # Balance replicas over replication-groups for each partition
        if any(b.inactive for b in self.brokers.itervalues()):
            self.log.error(
                "Impossible to rebalance replication groups because of inactive "
                "brokers."
            )
            raise RebalanceError(
                "Impossible to rebalance replication groups because of inactive "
                "brokers"
            )

        for partition in self.partitions.itervalues():
            self._rebalance_partition(partition)

        # Balance partition-count over replication-groups
        self._rebalance_groups_partition_cnt()

    def decommission_brokers(self, broker_ids):
        """Decommission a list of brokers trying to keep the replication group
        the brokers belong to balanced.

        :param broker_ids: list of string representing valid broker ids in the cluster
        :raises: InvalidBrokerIdError when the id is invalid.
        """
        groups = set()
        for b_id in broker_ids:
            try:
                broker = self.brokers[b_id]
            except KeyError:
                self.log.error("Invalid broker id %s.", b_id)
                # Raise an error for now. As alternative we may ignore the
                # invalid id and continue with the others.
                raise InvalidBrokerIdError(
                    "Broker id {} does not exist in cluster".format(b_id),
                )
            broker.mark_decommissioned()
            groups.add(broker.replication_group)

        for group in groups:
            self._decommission_brokers_in_group(group)

    def _decommission_brokers_in_group(self, group):
        """Decommission the marked brokers of a group."""
        try:
            group.rebalance_brokers()
        except EmptyReplicationGroupError:
            self.log.warning("No active brokers left in replication group %s", group)
        for broker in group.brokers:
            if broker.decommissioned and not broker.empty():
                # In this case we need to reassign the remaining partitions
                # to other replication groups
                self.log.info(
                    "Broker %s can't be decommissioned within the same "
                    "replication group %s. Moving partitions to other "
                    "replication groups.",
                    broker,
                    broker.replication_group,
                )
                self._force_broker_decommission(broker)
                # Broker should be empty now
                if not broker.empty():
                    # Decommission may be impossible if there are not enough
                    # brokers to redistributed the replicas.
                    self.log.error(
                        "Could not decommission broker %s. "
                        "Partitions %s cannot be reassigned.",
                        broker,
                        broker.partitions,
                    )
                    raise BrokerDecommissionError("Broker decommission failed.")

    def _force_broker_decommission(self, broker):
        available_groups = [
            rg for rg in self.rgs.itervalues()
            if rg is not broker.replication_group
        ]

        for partition in broker.partitions.copy():  # partitions set changes during loop
            groups = sorted(
                available_groups,
                key=lambda x: x.count_replica(partition),
            )
            for group in groups:
                self.log.debug(
                    "Try to move partition: %s from broker %s to "
                    "replication group %s",
                    partition,
                    broker,
                    broker.replication_group,
                )
                try:
                    group.acquire_partition(partition, broker)
                    break
                except NotEligibleGroupError:
                    pass

    def replace_broker(self, source_id, dest_id):
        """Move all partitions in source broker to destination broker.

        :param source_id: source broker-id
        :param dest_id: destination broker-id
        :raises: InvalidBrokerIdError, when either of given broker-ids is invalid.
        """
        try:
            source = self.brokers[source_id]
            dest = self.brokers[dest_id]
            # Move all partitions from source to destination broker
            for partition in source.partitions.copy():  # Partitions set changes
                # We cannot move partition directly since that re-orders the
                # replicas for the partition
                source.partitions.remove(partition)
                dest.partitions.add(partition)
                # Replace broker in replica
                partition.replace(source, dest)
        except KeyError as e:
            self.log.error("Invalid broker id %s.", e[0])
            raise InvalidBrokerIdError(
                "Broker id {} does not exist in cluster".format(e[0])
            )

    def _rebalance_partition(self, partition):
        """Rebalance replication group for given partition."""
        # Separate replication-groups into under and over replicated
        total = sum(
            group.count_replica(partition)
            for group in self.rgs.itervalues()
        )
        over_replicated_rgs, under_replicated_rgs = separate_groups(
            self.rgs.values(),
            lambda g: g.count_replica(partition),
            total,
        )
        # Move replicas from over-replicated to under-replicated groups
        while under_replicated_rgs and over_replicated_rgs:
            # Decide source and destination group
            rg_source = self._elect_source_replication_group(
                over_replicated_rgs,
                partition,
            )
            rg_destination = self._elect_dest_replication_group(
                rg_source.count_replica(partition),
                under_replicated_rgs,
                partition,
            )
            if rg_source and rg_destination:
                # Actual movement of partition
                self.log.debug(
                    'Moving partition {p_name} from replication-group '
                    '{rg_source} to replication-group {rg_dest}'.format(
                        p_name=partition.name,
                        rg_source=rg_source.id,
                        rg_dest=rg_destination.id,
                    ),
                )
                rg_source.move_partition(rg_destination, partition)
            else:
                # Groups balanced or cannot be balanced further
                break
            # Re-compute under and over-replicated replication-groups
            over_replicated_rgs, under_replicated_rgs = separate_groups(
                self.rgs.values(),
                lambda g: g.count_replica(partition),
                total,
            )

    def _elect_source_replication_group(
        self,
        over_replicated_rgs,
        partition,
    ):
        """Decide source replication-group based as group with highest replica
        count.
        """
        return max(
            over_replicated_rgs,
            key=lambda rg: rg.count_replica(partition),
        )

    def _elect_dest_replication_group(
        self,
        replica_count_source,
        under_replicated_rgs,
        partition,
    ):
        """Decide destination replication-group based on replica-count."""
        min_replicated_rg = min(
            under_replicated_rgs,
            key=lambda rg: rg.count_replica(partition),
        )
        # Locate under-replicated replication-group with lesser
        # replica count than source replication-group
        if min_replicated_rg.count_replica(partition) < replica_count_source - 1:
            return min_replicated_rg
        return None

    # Re-balancing partition count across brokers
    def _rebalance_groups_partition_cnt(self):
        """Re-balance partition-count across replication-groups.

        Algorithm:
        The key constraint is not to create any replica-count imbalance while
        moving partitions across replication-groups.
        1) Divide replication-groups into over and under loaded groups in terms
           of partition-count.
        2) For each over-loaded replication-group, select eligible partitions
           which can be moved to under-replicated groups. Partitions with greater
           than optimum replica-count for the group have the ability to donate one
           of their replicas without creating replica-count imbalance.
        3) Destination replication-group is selected based on minimum partition-count
           and ability to accept one of the eligible partition-replicas.
        4) Source and destination brokers are selected based on :-
            * their ability to donate and accept extra partition-replica respectively.
            * maximum and minimum partition-counts respectively.
        5) Move partition-replica from source to destination-broker.
        6) Repeat steps 1) to 5) until groups are balanced or cannot be balanced further.
        """
        # Segregate replication-groups based on partition-count
        total_elements = sum(len(rg.partitions) for rg in self.rgs.itervalues())
        over_loaded_rgs, under_loaded_rgs = separate_groups(
            self.rgs.values(),
            lambda rg: len(rg.partitions),
            total_elements,
        )
        if over_loaded_rgs and under_loaded_rgs:
            self.log.info(
                'Over-loaded replication-groups {over_loaded}, under-loaded '
                'replication-groups {under_loaded} based on partition-count'
                .format(
                    over_loaded=[rg.id for rg in over_loaded_rgs],
                    under_loaded=[rg.id for rg in under_loaded_rgs],
                )
            )
        else:
            self.log.info('Replication-groups are balanced based on partition-count.')
            return

        # Get optimal partition-count per replication-group
        opt_partition_cnt, _ = compute_optimum(
            len(self.rgs),
            total_elements,
        )
        # Balance replication-groups
        for over_loaded_rg in over_loaded_rgs:
            for under_loaded_rg in under_loaded_rgs:
                # Filter unique partition with replica-count > opt-replica-count
                # in over-loaded-rgs and <= opt-replica-count in under-loaded-rgs
                eligible_partitions = set(filter(
                    lambda partition:
                    over_loaded_rg.count_replica(partition) >
                    len(partition.replicas) // len(self.rgs) and
                    under_loaded_rg.count_replica(partition) <=
                    len(partition.replicas) // len(self.rgs),
                    over_loaded_rg.partitions,
                ))
                # Move all possible partitions
                for eligible_partition in eligible_partitions:
                    over_loaded_rg.move_partition_replica(
                        under_loaded_rg,
                        eligible_partition,
                    )
                    # Move to next replication-group if either of the groups got
                    # balanced, otherwise try with next eligible partition
                    if (len(under_loaded_rg.partitions) == opt_partition_cnt or
                            len(over_loaded_rg.partitions) == opt_partition_cnt):
                        break
                if len(over_loaded_rg.partitions) == opt_partition_cnt:
                    # Move to next over-loaded replication-group if balanced
                    break

    # Re-balancing partition count across brokers
    def rebalance_brokers(self):
        """Rebalance partition-count across brokers within each replication-group."""
        for rg in self.rgs.itervalues():
            rg.rebalance_brokers()

    # Re-balancing leaders
    def rebalance_leaders(self):
        """Re-order brokers in replicas such that, every broker is assigned as
        preferred leader evenly.
        """
        opt_leader_cnt = len(self.partitions) // len(self.brokers)
        # Balanced brokers transfer leadership to their under-balanced followers
        self.rebalancing_non_followers(opt_leader_cnt)

    def rebalancing_non_followers(self, opt_cnt):
        """Transfer leadership to any under-balanced followers on the pretext
        that they remain leader-balanced or can be recursively balanced through
        non-followers (followers of other leaders).

        Context:
        Consider a graph G:
        Nodes: Brokers (e.g. b1, b2, b3)
        Edges: From b1 to b2 s.t. b1 is a leader and b2 is its follower
        State of nodes:
            1. Over-balanced/Optimally-balanced: (OB)
                if leadership-count(broker) >= opt-count
            2. Under-balanced (UB) if leadership-count(broker) < opt-count
            leader-balanced: leadership-count(broker) is in [opt-count, opt-count+1]

        Algorithm:
            1. Use Depth-first-search algorithm to find path between
            between some UB-broker to some OB-broker.
            2. If path found, update UB-broker and delete path-edges (skip-partitions).
            3. Continue with step-1 until all possible paths explored.
        """
        under_brokers = filter(
            lambda b: b.count_preferred_replica() < opt_cnt,
            self.brokers.itervalues(),
        )
        if under_brokers:
            skip_brokers, skip_partitions = [], []
            for broker in under_brokers:
                skip_brokers.append(broker)
                broker.request_leadership(opt_cnt, skip_brokers, skip_partitions)

        over_brokers = filter(
            lambda b: b.count_preferred_replica() > opt_cnt + 1,
            self.brokers.itervalues(),
        )
        # Any over-balanced brokers tries to donate their leadership to followers
        if over_brokers:
            skip_brokers, used_edges = [], []
            for broker in over_brokers:
                skip_brokers.append(broker)
                broker.donate_leadership(opt_cnt, skip_brokers, used_edges)

    def update_cluster_topology(self, assignment):
        """Modify the cluster-topology with given assignment.

        Change the replica set of partitions as in given assignment.

        :param assignment: dict representing actions to be used to update the current
        cluster-topology
        :raises: InvalidBrokerIdError when broker-id is invalid
        :raises: InvalidPartitionError when partition-name is invalid
        """
        try:
            for partition_name, replica_ids in assignment.iteritems():
                try:
                    new_replicas = [self.brokers[b_id] for b_id in replica_ids]
                except KeyError:
                    self.log.error(
                        "Invalid replicas %s for topic-partition %s-%s.",
                        ', '.join([str(id) for id in replica_ids]),
                        partition_name[0],
                        partition_name[1],
                    )
                    raise InvalidBrokerIdError(
                        "Invalid replicas {0}.".format(
                            ', '.join([str(id) for id in replica_ids])
                        ),
                    )
                try:
                    partition = self.partitions[partition_name]
                    old_replicas = [broker for broker in partition.replicas]

                    # Remove old partitions from broker
                    # This also updates partition replicas
                    for broker in old_replicas:
                        broker.remove_partition(partition)

                    # Add new partition to brokers
                    for broker in new_replicas:
                        broker.add_partition(partition)
                except KeyError:
                    self.log.error(
                        "Invalid topic-partition %s-%s.",
                        partition_name[0],
                        partition_name[1],
                    )
                    raise InvalidPartitionError(
                        "Invalid topic-partition {0}-{1}."
                        .format(partition_name[0], partition_name[1]),
                    )
        except KeyError:
            self.log.error("Could not parse given assignment {0}".format(assignment))
            raise



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import

from kafka_utils.kafka_check import status_code
from kafka_utils.kafka_check.commands.command import get_broker_id
from kafka_utils.kafka_check.commands.command import KafkaCheckCmd
from kafka_utils.util.metadata import get_topic_partition_metadata


# This check will look on lines with that error-code in error field
# from kafka metadata response.
REPLICA_NOT_AVAILABLE_ERROR = 9


class UnderReplicatedCmd(KafkaCheckCmd):

    def build_subparser(self, subparsers):
        subparser = subparsers.add_parser(
            'under_replicated',
            description='Check under replicated partitions for all '
                        'brokers in cluster.',
            help='This command will sum all under replicated partitions '
                 'for each broker if any. It will query jolokia port for '
                 'receive this data.',
        )
        subparser.add_argument(
            '--first-broker-only',
            action='store_true',
            help='If this parameter is specified, it will do nothing and succeed '
                 'on not first brokers from Kafka cluster. Set --broker-id to -1 '
                 'to read broker-id from --data-path. Default: %(default)s',
        )

        return subparser

    def run_command(self):
        """Under_replicated command, checks number of under replicated partitions for
        all brokers in the Kafka cluster."""
        broker_list = self.zk.get_brokers()

        if self.args.first_broker_only:
            if self.args.broker_id is None:
                return status_code.WARNING, 'Broker id is not specified'

            if not _check_run_on_first_broker(broker_list, self.args.broker_id, self.args.data_path):
                return status_code.OK, 'Provided broker is not the first in broker-list.'

        under_replicated = _get_under_replicated(
            self.cluster_config.broker_list
        )

        if not under_replicated:
            return status_code.OK, 'No under replicated partitions.'
        else:
            if self.args.verbose:
                for (topic, partition) in under_replicated:
                    print('{topic}:{partition}'.format(
                        topic=topic,
                        partition=partition,
                    ))

            msg = "{under_replicated} under replicated partitions.".format(
                under_replicated=len(under_replicated),
            )
            return status_code.CRITICAL, msg


def _check_run_on_first_broker(broker_list, broker_id, data_path):
    """Returns true if the first broker in broker_list the same as in args."""
    broker_id = broker_id if broker_id != -1 else get_broker_id(data_path)

    first_broker_id, _ = min(broker_list.items())

    return broker_id == first_broker_id


def _process_topic_partition_metadata(topic_partitions_metadata):
    """Return set with under replicated partitions."""
    under_replicated = set()
    for partitions in topic_partitions_metadata.values():
        for metadata in partitions.values():
            if int(metadata.error) == REPLICA_NOT_AVAILABLE_ERROR:
                under_replicated.add((metadata.topic, metadata.partition))

    return under_replicated


def _get_under_replicated(broker_list):
    """Requests kafka-broker for metadata info for topics.
    Then checks if topic-partition is under replicated and there are not enough
    replicas in sync. Returns set of under replicated partitions.

    :param dictionary broker_list: dictionary with brokers information, broker_id is key
    :returns set: with under replicated partitions

        * set: { (topic, partition), ... }
    """
    metadata = get_topic_partition_metadata(broker_list)

    return _process_topic_partition_metadata(metadata)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
from mock import Mock
from mock import sentinel

from kafka_utils.kafka_cluster_manager.cluster_info.partition import Partition
from kafka_utils.kafka_cluster_manager.cluster_info.topic import Topic


class TestTopic(object):

    @pytest.fixture
    def topic(self):
        return Topic('t0', 2, set([sentinel.p1, sentinel.p2]))

    def test_id(self, topic):
        assert topic.id == 't0'

    def test_replication_factor(self, topic):
        assert topic.replication_factor == 2

    def test_partitions(self, topic):
        assert topic.partitions == set([sentinel.p1, sentinel.p2])

    def test_add_partition(self):
        mock_partitions = set([
            Mock(
                spec=Partition,
                replicas=[sentinel.r1, sentinel.r2],
            ),
            Mock(
                spec=Partition,
                replicas=[sentinel.r4, sentinel.r3],
            ),
        ])
        topic = Topic('t0', 2, mock_partitions)
        new_partition = Mock(spec=Partition, replicas=[sentinel.r2])
        topic.add_partition(new_partition)
        assert topic.partitions == mock_partitions | set([new_partition])



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock
import pytest
import requests
from requests.exceptions import RequestException

from kafka_utils.kafka_rolling_restart import main


@mock.patch.object(main.FuturesSession, 'get', autospec=True)
def test_read_cluster_value_partitions(mock_get):
    response = mock.Mock(status_code=200, spec=requests.Response)
    response.json.return_value = {'value': 1}

    request = mock_get.return_value
    request.result.return_value = response

    p, b = main.read_cluster_status(["host1", "host2", "host3"], 80, "jolokia")

    assert p == 3   # 3 missing partitions
    assert b == 0   # 0 missing brokers


@mock.patch.object(main.FuturesSession, 'get', autospec=True)
def test_read_cluster_value_exit(mock_get):
    response = mock.Mock(status_code=404, spec=requests.Response)

    request = mock_get.return_value
    request.result.return_value = response

    with pytest.raises(SystemExit):
        p, b = main.read_cluster_status(["host1"], 80, "jolokia")


@mock.patch.object(main.FuturesSession, 'get', autospec=True)
def test_read_cluster_value_no_key(mock_get):
    response = mock.Mock(status_code=200, spec=requests.Response)
    response.json.return_value = {'wrong_key': 1}

    request = mock_get.return_value
    request.result.return_value = response

    p, b = main.read_cluster_status(["host1"], 80, "jolokia")

    assert p == 0   # 0 missing partitions
    assert b == 1   # 1 missing brokers


@mock.patch.object(main.FuturesSession, 'get', autospec=True)
def test_read_cluster_value_server_down(mock_get):
    request = mock_get.return_value
    request.result.side_effect = RequestException

    p, b = main.read_cluster_status(["host1"], 80, "jolokia")

    assert p == 0   # 0 missing partitions
    assert b == 1   # 1 missing brokers


def read_cluster_state_values(first_part, repeat):
    for value in first_part:
        yield value
    while True:
        yield repeat


@mock.patch.object(
    main,
    'read_cluster_status',
    side_effect=read_cluster_state_values([(100, 1), (0, 1), (100, 0)], (0, 0)),
    autospec=True,
)
@mock.patch('time.sleep', autospec=True)
def test_wait_for_stable_cluster_success(mock_sleep, mock_read):
    main.wait_for_stable_cluster([], 1, "", 5, 3, 100)

    assert mock_read.call_count == 6
    assert mock_sleep.mock_calls == [mock.call(5)] * 5


@mock.patch.object(
    main,
    'read_cluster_status',
    side_effect=read_cluster_state_values([], (100, 0)),
    autospec=True,
)
@mock.patch('time.sleep', autospec=True)
def test_wait_for_stable_cluster_timeout(mock_sleep, mock_read):
    with pytest.raises(main.WaitTimeoutException):
        main.wait_for_stable_cluster([], 1, "", 5, 3, 100)

    assert mock_read.call_count == 21
    assert mock_sleep.mock_calls == [mock.call(5)] * 20
